﻿/*
 * Mixer Unity SDK
 *
 * Copyright (c) Microsoft Corporation
 * All rights reserved.
 *
 * MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the "Software"), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
using System.Collections.Generic;

namespace Microsoft.Mixer
{
    /// <summary>
    /// Represents an interactive scene. These scenes are configured using
    /// Interactive Studio.
    /// </summary>
#if !WINDOWS_UWP
    [System.Serializable]
#endif
    public class InteractiveScene
    {
        /// <summary>
        /// Unique string identifier for this scene
        /// </summary>
        public string SceneID
        {
            get;
            internal set;
        }

        /// <summary>
        /// Retrieve a List of all of the buttons in the scene. May be empty.
        /// </summary>
        public IList<InteractiveButtonControl> Buttons
        {
            get
            {
                List<InteractiveButtonControl> buttonsInScene = new List<InteractiveButtonControl>();
                List<InteractiveButtonControl> allButtons = InteractivityManager.SingletonInstance.Buttons as List<InteractiveButtonControl>;
                foreach (InteractiveButtonControl button in allButtons)
                {
                    if (button._sceneID == SceneID)
                    {
                        buttonsInScene.Add(button);
                    }
                }
                return buttonsInScene;
            }
        }

        /// <summary>
        /// Retrieve a List of all of the joysticks in the scene. May be empty.
        /// </summary>
        public IList<InteractiveJoystickControl> Joysticks
        {
            get
            {
                List<InteractiveJoystickControl> joysticksInScene = new List<InteractiveJoystickControl>();
                List<InteractiveJoystickControl> allJoysticks = InteractivityManager.SingletonInstance.Joysticks as List<InteractiveJoystickControl>;
                foreach (InteractiveJoystickControl joystick in allJoysticks)
                {
                    if (joystick._sceneID == SceneID)
                    {
                        joysticksInScene.Add(joystick);
                    }
                }
                return joysticksInScene;
            }
        }

        /// <summary>
        /// Retrieve a list of IDs of all of the groups assigned to the scene. May be empty.
        /// </summary>
        public IList<InteractiveGroup> Groups
        {
            get
            {
                List<InteractiveGroup> groupsAssignedToScene = new List<InteractiveGroup>();
                List<InteractiveGroup> allGroups = InteractivityManager.SingletonInstance.Groups as List<InteractiveGroup>;
                foreach (InteractiveGroup group in allGroups)
                {
                    if (group.SceneID == SceneID)
                    {
                        groupsAssignedToScene.Add(group);
                    }
                }
                return groupsAssignedToScene;
            }
        }

        /// <summary>
        /// Retrieves a reference to the specified button, if it exists.
        /// </summary>
        public InteractiveButtonControl GetButton(string controlID)
        {
            return InteractivityManager.SingletonInstance.GetButton(controlID);
        }

        /// <summary>
        /// Retrieve a vector of all of the joysticks in the scene. May be empty.
        /// </summary>
        public InteractiveJoystickControl GetJoystick(string controlID)
        {
            return InteractivityManager.SingletonInstance.GetJoystick(controlID);
        }

        internal string _etag;

        internal InteractiveScene(string sceneID = "", string newEtag = "")
        {
            SceneID = sceneID;
            _etag = newEtag;
        }
    }
}
